{********************************************************}
{       TVirtComp component for Delphi 2,3,4             }
{                     Version 1.3                        }
{       Copyright(c) Alexander Baranovsky,1999           }
{              mailto: ab@iai.donetsk.ua                 }
{********************************************************}

unit Virtcomp;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, dsgnintf;
const
  LibName = 'virtcomp.dll';
  IniName = 'virt.ini';

  MaxUserLibraryCount = 30;
  MaxUserFunctionCount = 200;
  MaxParams = 100;
type
  TVirtComp = class;

  TBeforeParsingHandler = procedure( var Sender : TVirtComp ) of object;
  TErrorHandler = procedure( Sender : TObject ) of object;

  TVirtState = ( stNULL, stINIT, stLOAD );

  TExternalFunc = procedure;

  TVirtComp = class(TComponent)
  private
    fH : Integer;
    FFont : TFont;
    fScript   : TStrings;
    fFileName : TFileName;
    fOnBeforeParsing : TBeforeParsingHandler;
    fOnCompileError,
    fOnRunError,
    fOnSaveBinProgError,
    fOnLoadBinProgError : TErrorHandler;
    fAssoc : TStringList;
    fState : TVirtState;
    fModeParse : Byte;
    fParams : array[ 1..MaxParams ] of Variant;
    fError, fErrorFile : ShortString;
    fErrorLine, fErrorPos : Integer;
    procedure SetFont( const Value : TFont );
    function  GetParam( const AIndex : Integer ) : Variant;
    procedure SetParam( const AIndex : Integer; const Value : Variant );
    procedure OpenVirt;
    procedure CloseVirt;
  protected
    procedure ShowError; virtual;
    procedure SetGlobals;
    function  ScriptExists : boolean;
    function  LoadScript : boolean;
    procedure ProcessRunError;
    function  GetIDOfVariable( const VarName : String ) : Integer;
    function  GetIDOfFunction( const FuncName : String ) : Integer;
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;
    procedure  Clear;
    function   SaveBinProg : boolean;
    function   LoadBinProg : boolean;
    function   Parse : boolean;
    function   Run : boolean;
    function   ExecFunction( const FuncName : String;
          const Params : array of Variant; var Res : Variant ) : boolean;
    procedure  AppFunction( const FuncName : String;
                            ParCount : Integer;
                            Proc : TExternalFunc );
    procedure  AppSystemFunction( const FuncName : String;
                                  ParCount : Integer;
                                  Proc : TExternalFunc );
    function   PopVar : Variant;
    function   PopVarAsString : String;
    function   PopIDOfVariable : Integer;
    procedure  PushVar( const V : Variant );
    function   GetVar( const VarName : String; var Res : Variant ) : boolean;
    function   GetVarAsString( const VarName : String;
                               var Res : ShortString ) : boolean;
    function   PutVar( const VarName : String; const Val : Variant ) : boolean;
    function   PutVarByID( ID : Integer; const Val : Variant ) : boolean;
    procedure  PushObject( const ClassName : String;
                           const Fields : array of Variant );
    procedure  SetVar( const VarName : String; const Val : Variant );
    procedure  GetLastError( var ErrorMessage, ErrorFile : ShortString;
                             var ErrorLine, ErrorPos : Integer );
    function   GetDllPath : String;
    property   Params[ const AIndex : Integer ] : Variant read GetParam write SetParam;
  published
    property   Font : TFont read FFont write SetFont;
    property   Script : TStrings read fScript write fScript;
    property   ScriptFileName : TFileName read fFileName write fFileName;
    property   OnBeforeParsing : TBeforeParsingHandler read fOnBeforeParsing write fOnBeforeParsing;
    property   OnCompileError : TErrorHandler read fOnCompileError write fOnCompileError;
    property   OnRunError : TErrorHandler read fOnRunError write fOnRunError;
    property   OnSaveBinProgError : TErrorHandler read fOnSaveBinProgError write fOnSaveBinProgError;
    property   OnLoadBinProgError : TErrorHandler read fOnLoadBinProgError write fOnLoadBinProgError;
  end;

  TFileNameEdit = class(TPropertyEditor)
  public
    function  GetAttributes : TPropertyAttributes; override;
    function  GetValue : string; override;
    procedure SetValue(const val : string); override;
    procedure Edit; override;
  end;

  TMemoEdit = class(TPropertyEditor)
  public
    function  GetAttributes : TPropertyAttributes; override;
    function  GetValue : string; override;
    procedure SetValue(const val : string); override;
    procedure Edit; override;
  end;

  TVirtCompEditor = class( TComponentEditor )
    procedure Edit; override;
  end;

  TProcedure = procedure;

  TUserFuncRec = record
    FuncName : String;
    ParamCount : Integer;
    FuncAddr : TProcedure;
  end;

procedure Register;

var
  RunVirtComp : TVirtComp;

  UserLibraryHandlers : array[ 1..MaxUserLibraryCount ]of Integer;
  UserLibraryCount : Integer;

  UserFunctions : array[ 1..MaxUserFunctionCount ] of TUserFuncRec;
  UserFunctionCount : Integer;

implementation

uses
  TypInfo;

const
  CError='Error';
  CConfirm='Confirm';
  CCouldNotFind='Could not find file ';
  CIncorrectParam='Incorrect VIRT parameter number';
  CFileExists='File %s already exists.'#13#10'Do you wish to replace it?';
  CCouldNotLoad='Could not load DLL ';
  CCannotCreate='Cannot create file ';
  CCannotLoad='Cannot load file ';
  CScriptFilenotFound='Script file %s not found';
  CCannotCompile='Cannot compile file ';
  CCompileError='Compile error : %s'#13#10'File : %s'#13#10'Line : %d'#13#10'%s';
  CFormCaption='Error : %s, File : %s';
  CReadOnly=' (ReadOnly at run-time)';
  CFunctionnotFound='Function %s is not found';
  CCannotView='Cannot view file ';
  CSaveChanges='Save changes to %s?';
  CUnregisteredDelphiClass = 'Unregistered Delphi class';

procedure _writeln; forward;

type

TVirtInit = function( Handle : THandle ) : Integer;
TVirtDone = procedure( H : Integer );
TVirtParse = function( H : Integer; const ProgName : String; ModeParse : Byte ) : boolean;
TVirtRun = function( H : Integer ) : boolean;
TVirtAppFunction = procedure( H : Integer; const FuncName : String;
                              ParamCount : Integer;
                              Proc : TExternalFunc );
TVirtAppSystemFunction = TVirtAppFunction;
TVirtSaveBinProg = function( H : Integer; const BinProgName : String ) : boolean;
TVirtLoadBinProg = function( H : Integer; const BinProgName : String ) : boolean;
TVirtGetIDOfFunction = function( const FuncName : String ) : Integer;
TVirtExecFunction = function( FuncAddr : Integer;
            const V : array of Variant; var R : Variant ) : boolean;
TVirtPopVar = function : Variant;
TVirtPopIDOfVariable = function : Integer;
TVirtPushVar = procedure( const V : Variant );
TVirtGetIDOfVariable = function( const VarName : String ) : Integer;
TVirtGetVar = function( VarAddr : Integer ) : Variant;
TVirtPutVar = procedure( VarAddr : Integer; const V : Variant );
TVirtPopVarAsString = function : ShortString;
TVirtGetError = function( H, What : Integer;
                          var Error : ShortString;
                          var ErrorFile : ShortString;
                          var ErrorLine : Integer;
                          var ErrorPos  : Integer ) : boolean;
TVirtSetRunComp = procedure( var InitNCurr : Integer );
TVirtPushObject = procedure( const ClassName : String;
                             const FieldsList : array of Variant );
TVirtSetVar = procedure( H : Integer; const VarName : String; const Value : Variant );
TVirtGetVarAsString = function( VarAddr : Integer ) : ShortString;
TVirtGetDllPath = function : ShortString;

var
    LibraryHandle : THandle;
    VirtCount : Integer;

    VirtInit  : TVirtInit;
    VirtParse : TVirtParse;
    VirtRun   : TVirtRun;
    VirtDone  : TVirtDone;
    VirtGetIDOfFunction : TVirtGetIDOfFunction;
    VirtExecFunction : TVirtExecFunction;
    VirtAppFunction : TVirtAppFunction;
    VirtAppSystemFunction : TVirtAppSystemFunction;
    VirtPopVar : TVirtPopVar;
    VirtPushVar : TVirtPushVar;
    VirtSaveBinProg : TVirtSaveBinProg;
    VirtLoadBinProg : TVirtLoadBinProg;
    VirtGetIDOfVariable : TVirtGetIDOfVariable;
    VirtGetVar : TVirtGetVar;
    VirtPopVarAsString : TVirtPopVarAsString;
    VirtPopIDOfVariable : TVirtPopIDOfVariable;
    VirtPutVar : TVirtPutVar;
    VirtGetError : TVirtGetError;
    VirtSetRunComp : TVirtSetRunComp;
    VirtPushObject : TVirtPushObject;
    VirtSetVar : TVirtSetVar;
    VirtGetVarAsString : TVirtGetVarAsString;
    VirtGetDllPath : TVirtGetDllPath;

function PutProperty( Instance : TObject; const PropName : String; const Value : Variant ) : boolean;
var
  P : PPropInfo;
begin
  result := false;
  P := GetPropInfo( Instance.ClassInfo, PropName );
  if P <> nil then
  begin
    result := true;
    case P.PropType^.Kind of
      tkLString : SetStrProp( Instance, P, Value );
      tkInteger : SetOrdProp( Instance, P, Value );
      tkFloat   : SetFloatProp( Instance, P, Value );
    else
      result := false;
    end;
  end;
end;

function GetProperty( Instance : TObject; const PropName : String; var Value : Variant ) : boolean;
var
  P : PPropInfo;
begin
  result := false;
  P := GetPropInfo( Instance.ClassInfo, PropName );
  if P <> nil then
  begin
    result := true;
    case P.PropType^.Kind of
      tkLString : Value := GetStrProp( Instance, P );
      tkInteger : Value := GetOrdProp( Instance, P );
      tkFloat   : Value := GetFloatProp( Instance, P );
    else
      result := false;
    end;
  end;
end;

procedure _Delphi_Init;
var
  Name  : String;
  C : TComponent;
begin
  Name := RunVirtComp.PopVar;    // Component's name

  if UpperCase( Name ) = 'SELF' then
    C := RunVirtComp.Owner
  else
    C := RunVirtComp.Owner.FindComponent( Name );

  RunVirtComp.PushVar( Integer(C) ); // push into stack
  RunVirtComp.PutVar( 'Delphi.Instance', Integer(C) ); // assign field
end;

procedure _Delphi_New;
var
  Name, ClassName : String;
  C : TComponent;
begin
  Name := RunVirtComp.PopVar;      //2th parameter
  ClassName := RunVirtComp.PopVar; //1th parameter

  if      UpperCase( ClassName ) = 'TBUTTON' then
    C := TButton.Create( RunVirtComp.Owner )
  else if UpperCase( ClassName ) = 'TLABEL' then
    C := TLabel.Create( RunVirtComp.Owner )
  else if UpperCase( ClassName ) = 'TEDIT' then
    C := TEdit.Create( RunVirtComp.Owner )
  else if UpperCase( ClassName ) = 'TMEMO' then
    C := TMemo.Create( RunVirtComp.Owner )
  else if UpperCase( ClassName ) = 'TPANEL' then
    C := TPanel.Create( RunVirtComp.Owner )
  else if UpperCase( ClassName ) = 'TLISTBOX' then
    C := TListBox.Create( RunVirtComp.Owner )
  else if UpperCase( ClassName ) = 'TCOMBOBOX' then
    C := TComboBox.Create( RunVirtComp.Owner )
  else if UpperCase( ClassName ) = 'TCHECKBOX' then
    C := TCheckBox.Create( RunVirtComp.Owner )
  else if UpperCase( ClassName ) = 'TRADIOBUTTON' then
    C := TRadioButton.Create( RunVirtComp.Owner )
  else
    raise Exception.Create( CUnregisteredDelphiClass );

  TWinControl( C ).Parent := TWinControl( RunVirtComp.Owner );
  C.Name := Name;

  RunVirtComp.PushObject( 'Delphi', ['Delphi.Instance', Integer(C)] );
end;

procedure _Delphi_IndexedItem;
var
  V : Variant;
  C, I : Integer;
  Item : TComponent;
  VType : Byte;
  S : String;
begin
  RunVirtComp.GetVar( 'Delphi.Instance', V );
  C := V;

  V := RunVirtComp.PopVar; // Index

  VType := TVarData( V ).VType;
  if VType = VARINTEGER then
  begin
    I := V;
    Item := TComponent( C ).Components[ I ];
  end
  else
  begin
    S := V;
    Item := TComponent( C ).FindComponent( S );
  end;

  RunVirtComp.PushObject( 'Delphi', ['Delphi.Instance', Integer(Item)] );
end;

procedure _Delphi_GetOwner;
var
  V : Variant;
  C : Integer;
  Owner : TComponent;
begin
  RunVirtComp.GetVar( 'Delphi.Instance', V );
  C := V;
  Owner := TComponent( C ).Owner;
  RunVirtComp.PushObject( 'Delphi', ['Delphi.Instance', Integer(Owner)] );
end;

procedure _Delphi_ComponentCount;
var
  C : Integer;
  V : Variant;
begin
  RunVirtComp.GetVar( 'Delphi.Instance', V );
  C := V;
  RunVirtComp.PushVar( TComponent( C ).ComponentCount );
end;

procedure _Delphi_PutProperty;
var
  V, InitPropValue : Variant;
  PropName : String;
  C : Integer;
begin
  RunVirtComp.GetVar( 'Delphi.Instance', V );
  C := V;
  InitPropValue := RunVirtComp.PopVar; // 2th parameter
  PropName      := RunVirtComp.PopVar; // 1th parameter

  if PutProperty( TComponent( C ), PropName, InitPropValue ) then
    RunVirtComp.PushVar(1) // return value is "true"
  else
    RunVirtComp.PushVar(0); // return value is "false"
end;

procedure _Delphi_GetProperty;
var
  V, RetPropValue : Variant;
  PropName : String;
  C, ID_RetPropValue : Integer;
begin
  RunVirtComp.GetVar( 'Delphi.Instance', V );
  C := V;
  ID_RetPropValue := RunVirtComp.PopIDOfVariable; // 2th parameter by reference
  PropName := RunVirtComp.PopVar; // 1th parameter

  if GetProperty( TComponent( C ), PropName, RetPropValue ) then
  begin
    RunVirtComp.PushVar(1); // return value is "true"
    RunVirtComp.PutVarByID( ID_RetPropValue, RetPropValue ); // return parameter by reference
  end
  else
    RunVirtComp.PushVar(0); // return value is "false"
end;

procedure _Delphi_GetParam;
var
  N : Integer;
  V : Variant;
begin
  N := RunVirtComp.PopVar;
  if ( N >=1 ) and ( N <= MaxParams ) then
    V := RunVirtComp.Params[ N ];
  RunVirtComp.PushVar( V );
end;

procedure _Delphi_PutParam;
var
  N : Integer;
  V : Variant;
begin
  V := RunVirtComp.PopVar; // parameter
  N := RunVirtComp.PopVar; // index
  if ( N >=1 ) and ( N <= MaxParams ) then
    RunVirtComp.Params[ N ] := V;
  RunVirtComp.PushVar(0);
end;

procedure ErrMessage( const S : String );
begin
  MessageBox( GetActiveWindow(), PChar(S), PChar(CError), MB_ICONEXCLAMATION or MB_OK );
end;

function PosKeyWord( const KeyWord, S : String ) : Integer;
var
  P : Integer;
begin
  result := Pos( KeyWord, S );

  if result > 0 then
  begin
    P := result + Length(KeyWord);
    if P > Length( S ) then Exit
    else
    if S[ P ] <> ' ' then
       result := 0;
  end;

  if result > 1 then
    if S[ result - 1 ] <> ' ' then
       result := 0;
end;

function AppUserLib( const LibName, SourceName : String ) : Integer;
type
  TExternalProc = procedure;
const
  WordFunc = 'FUNCTION';
  WordImpl = 'IMPLEMENTATION';
var
  H : THandle;
  T : TextFile;
  S, FuncName, ClassName : String;
  P, ComaCount, ParamCount, PosComment : Integer;
  FuncAddr : TExternalFunc;
begin
  result := 0;

  ClassName := '';

  if not FileExists( SourceName ) then
  begin
    ErrMessage( CCouldNotFind + SourceName );
    Exit;
  end;

  H := LoadLibrary( PChar(LibName) );

  result := H;

  if H > 0 then
  begin
    AssignFile( T, SourceName );
    Reset( T );
    repeat
      Readln( T, S );
      S := UpperCase( S );

      PosComment := Pos( '--', S );
      if PosComment > 0 then
        S := Copy( S, 1, PosComment - 1 );

      P := PosKeyWord( WordImpl, S );
      if P > 0 then
      begin
        Inc( P, Length(WordImpl) );
        while S[P] = ' ' do Inc( P );
        ClassName := '';
        while ( S[P] <> ' ' ) and ( P <= Length(S) ) do
        begin
          ClassName := ClassName + S[P];
          Inc( P );
        end;
      end;

      if ( ClassName <> '' ) and ( PosKeyWord( 'END', S ) > 0 ) then
        ClassName := '';

      P := PosKeyWord( WordFunc, S );

      if ( P > 0 ) and ( PosKeyWord( 'EXTERNAL', S ) > 0 ) then
      begin
        Inc( P, Length(WordFunc) );
        while S[P] = ' ' do Inc( P );
        FuncName := '';
        while S[P] <> '(' do
        begin
          FuncName := FuncName + S[P];
          Inc( P );
        end;
        ComaCount := 0;
        for P:=1 to Length(S) do
          if S[P] = ',' then Inc( ComaCount );
        if ComaCount = 0 then
          ParamCount := 0
        else
          ParamCount := ComaCount + 1;

        FuncAddr := TExternalFunc( GetProcAddress( H, PChar(FuncName) ) );
        if Assigned( FuncAddr ) then
        begin
          Inc( UserFunctionCount );
          UserFunctions[ UserFunctionCount ].ParamCount := ParamCount;
          UserFunctions[ UserFunctionCount ].FuncAddr := FuncAddr;

          if ClassName = '' then
            UserFunctions[ UserFunctionCount ].FuncName := FuncName
          else
            UserFunctions[ UserFunctionCount ].FuncName := ClassName +
                                                           '.' + FuncName;
        end;
      end;
    until EOF( T );
    CloseFile( T );
  end
  else
    ErrMessage( CCouldNotLoad + LibName );
end;

procedure LoadUserLibraries;
var
  T : TextFile;
  S, NameDLL, NameW : String;
  Found, EndOfList : boolean;
  H : Integer;
begin
  if not FileExists( IniName ) then Exit;

  AssignFile( T, IniName );
  Reset( T );

  try
    Found := false;
    repeat
      Readln( T, S );
      S := UpperCase( S );
      if Pos( '[PLUG]', S ) > 0 then
        Found := true;
    until EOF( T ) or Found;

    if Found and ( not EOF(T) ) then
    begin
      repeat
        Readln( T, S );
        EndOfList := Pos( '[', S ) > 0;
        if not EndOfList then
        begin
          NameDLL := S + '.DLL';
          NameW   := S + '.W';
          if FileExists( NameDLL ) then
          begin
            if FileExists( NameW ) then
            try
              H := AppUserLib( NameDLL, NameW );
              if H > 0 then
              begin
                Inc( UserLibraryCount );
                UserLibraryHandlers[ UserLibraryCount ] := H;
              end;
            finally
            end;
          end;
        end;
      until EOF( T ) or EndOfList;
    end;
  finally
    CloseFile( T );
  end;
end;

constructor TVirtComp.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );

  fH := 0;

  FModeParse := 0;

  FFont := TFont.Create;
  FFont.Name := 'Courier new';
  FFont.Size := 10;

  if not ( csDesigning in ComponentState ) then
    OpenVirt;
end;

destructor TVirtComp.Destroy;
begin
  if not ( csDesigning in ComponentState ) then
    CloseVirt;

  FFont.Free;

  inherited Destroy;
end;

procedure TVirtComp.Clear;
begin
  fState := StINIT;
  fAssoc.Free;
  fAssoc := TStringList.Create;
end;

procedure TVirtComp.SetFont( const Value : TFont );
begin
  if Value <> nil then FFont.Assign( Value );
end;

procedure TVirtComp.SetParam( const AIndex : Integer; const Value : Variant );
begin
  if ( AIndex >= 1 ) and ( AIndex <= MaxParams ) then
    fParams[ AIndex ] := Value
  else
    raise Exception.Create( CIncorrectParam );
end;

function TVirtComp.GetParam( const AIndex : Integer ) : Variant;
begin
  if ( AIndex >= 1 ) and ( AIndex <= MaxParams ) then
    result := fParams[ AIndex ]
  else
    raise Exception.Create( CIncorrectParam );
end;

procedure TVirtComp.OpenVirt;
var
  I : Integer;
begin
  if VirtCount = 0 then
  begin
    LibraryHandle := LoadLibrary( LibName );
    if LibraryHandle > 0 then
    begin
      @VirtInit := GetProcAddress( LibraryHandle, 'VIRTINIT' );
      @VirtParse := GetProcAddress( LibraryHandle, 'VIRTPARSE' );
      @VirtRun := GetProcAddress( LibraryHandle, 'VIRTRUN' );
      @VirtDone := GetProcAddress( LibraryHandle, 'VIRTDONE' );
      @VirtGetIDOfFunction := GetProcAddress( LibraryHandle, 'VIRTGETIDOFFUNCTION'  );
      @VirtExecFunction := GetProcAddress( LibraryHandle, 'VIRTEXECFUNCTION' );
      @VirtAppFunction := GetProcAddress( LibraryHandle, 'VIRTAPPFUNCTION' );
      @VirtAppSystemFunction := GetProcAddress( LibraryHandle, 'VIRTAPPSYSTEMFUNCTION' );
      @VirtPopVar := GetProcAddress( LibraryHandle, 'VIRTPOPVAR' );
      @VirtPushVar := GetProcAddress( LibraryHandle, 'VIRTPUSHVAR' );
      @VirtSaveBinProg := GetProcAddress( LibraryHandle, 'VIRTSAVEBINPROG' );
      @VirtLoadBinProg := GetProcAddress( LibraryHandle, 'VIRTLOADBINPROG' );
      @VirtGetIDOfVariable := GetProcAddress( LibraryHandle, 'VIRTGETIDOFVARIABLE' );
      @VirtGetVar := GetProcAddress( LibraryHandle, 'VIRTGETVAR' );
      @VirtPopVarAsString := GetProcAddress( LibraryHandle, 'VIRTPOPVARASSTRING' );
      @VirtPopIDOfVariable := GetProcAddress( LibraryHandle, 'VIRTPOPIDOFVARIABLE' );
      @VirtPutVar := GetProcAddress( LibraryHandle, 'VIRTPUTVAR' );
      @VirtGetError := GetProcAddress( LibraryHandle, 'VIRTGETERROR' );
      @VirtSetRunComp := GetProcAddress( LibraryHandle, 'VIRTSETRUNCOMP' );
      @VirtPushObject := GetProcAddress( LibraryHandle, 'VIRTPUSHOBJECT' );
      @VirtSetVar := GetProcAddress( LibraryHandle, 'VIRTSETVAR' );
      @VirtGetVarAsString := GetProcAddress( LibraryHandle, 'VIRTGETVARASSTRING' );
      @VirtGetDllPath := GetProcAddress( LibraryHandle, 'VIRTGETDLLPATH' );
    end
    else
    begin
      ErrMessage( CCouldNotLoad + LibName );
      Exit;
    end;

    LoadUserLibraries;
  end;

  fH := VirtInit( LibraryHandle );

  AppFunction( 'writeln', 1, _writeln );
  AppFunction( 'Delphi.Init', 1, _Delphi_Init );
  AppFunction( 'Delphi.PutProperty', 2, _Delphi_PutProperty );
  AppFunction( 'Delphi.GetProperty', 2, _Delphi_GetProperty );
  AppFunction( 'Delphi.[]', 1, _Delphi_IndexedItem );
  AppFunction( 'Delphi.ComponentCount', 0, _Delphi_ComponentCount );
  AppFunction( 'Delphi.GetOwner', 0, _Delphi_GetOwner );
  AppFunction( 'Delphi.New', 2, _Delphi_New );
  AppFunction( 'Delphi.GetParam', 1, _Delphi_GetParam );
  AppFunction( 'Delphi.PutParam', 2, _Delphi_PutParam );

  for I:=1 to UserFunctionCount do
    with UserFunctions[ I ] do
      AppFunction( FuncName, ParamCount, FuncAddr );

  fState := stINIT;
  fAssoc := TStringList.Create;

  Inc( VirtCount );
end;

procedure TVirtComp.CloseVirt;
begin
  Dec( VirtCount );
  if VirtCount = 0 then
  begin
    if LibraryHandle > 0 then
    begin
      VirtDone( fH );
      FreeLibrary( LibraryHandle );
      LibraryHandle := 0;
    end;

    while UserLibraryCount > 0 do
    begin
      if UserLibraryHandlers[ UserLibraryCount ] > 0 then
        FreeLibrary( UserLibraryHandlers[ UserLibraryCount ] );
      Dec( UserLibraryCount );
    end;
  end;

  fState := stNULL;
  fAssoc.Free;
end;

function TVirtComp.PopVar : Variant;
begin
  Result := VirtPopVar;
end;

function TVirtComp.PopIDOfVariable : Integer;
begin
  Result := VirtPopIDOfVariable;
end;

function TVirtComp.PopVarAsString : String;
begin
  Result := VirtPopVarAsString;
end;

procedure  TVirtComp.PushVar( const V : Variant );
begin
  VirtPushVar( V );
end;

procedure  TVirtComp.PushObject( const ClassName : String;
                                 const Fields : array of Variant );
begin
  VirtPushObject( ClassName, Fields );
end;

function TVirtComp.SaveBinProg : boolean;
begin
  Result := true;
  if ScriptExists then
  begin
    if fState <> stLOAD then
      Result := Parse;
    if Result then
      if VirtSaveBinProg( fH, ScriptFileName + 'b' ) then Exit;
  end;

  Result := false;

  if Assigned( fOnSaveBinProgError ) then
    fOnSaveBinProgError( Self )
  else
    ErrMessage( CCannotCreate + ScriptFileName + 'b' );
end;

function TVirtComp.LoadBinProg : boolean;
begin
  if ScriptExists then
  begin
    Result := VirtLoadBinProg( fH, ScriptFileName );

    if Result then
    begin
      fState := stLOAD;
      Exit;
    end;
  end
  else
    Result := false;

  if Assigned( fOnLoadBinProgError ) then
    fOnLoadBinProgError( Self )
  else
    ErrMessage( CCannotLoad + ScriptFileName );
end;

function TVirtComp.ScriptExists : boolean;
begin
  result := FileExists( ScriptFileName );
  if not result then
    ErrMessage( Format(CScriptFilenotFound, [ScriptFileName]) );
end;

function FindLine( N : String; P : Integer ) : String;
var
  S : String;
  I : Integer;
  T : TextFile;
begin
  result := '';

  if not FileExists( N ) then Exit;

  AssignFile( T, N );
  Reset( T );

  try
    I := 0;
    repeat
      Inc( I );
      Readln( T, S );
      if I = P then
      begin
        result := S;
        Exit;
      end;
    until EOF( T );
  finally
    Close( T );
  end;
end;

function SaveNonameMemo( Memo : TMemo ) : String;
begin
  result := 'noname.w';
  with TSaveDialog.Create( nil ) do
    try
      Filter := 'VIRT files (*.w)|*.W' +
                '|VIRT source files (*.w)|*.W';
      if Execute then
      begin
        result := ExtractFileName( FileName );
        if Pos( '.', result ) = 0 then
          result := result + '.w';

        if FileExists( result ) then
        begin
          if MessageBox( GetActiveWindow(), PChar(Format(CFileExists, [result])),
            PChar(CConfirm), MB_ICONQUESTION or MB_YESNO ) = IDYES then
               Memo.Lines.SaveToFile( result );
        end
        else
          Memo.Lines.SaveToFile( result );
      end;
    finally
      Free;
    end;
end;

function TVirtComp.Parse : boolean;
begin
  if ScriptExists then
  begin
    if UpperCase( ExtractFileExt( ScriptFileName )) = '.WB' then
    begin
      Result := false;
      ErrMessage( CCannotCompile + ScriptFileName );
      Exit;
    end
    else
    begin
      if Assigned( fOnBeforeParsing ) then
         fOnBeforeParsing( Self );
      Result := VirtParse( fH, ScriptFileName, fModeParse );
    end;
  end
  else
    Result := false;

  if Result then
  begin
    fState := stLOAD;
    Exit;
  end;

  if Assigned( fOnCompileError ) then
    fOnCompileError( Self )
  else
  begin
    VirtGetError( fH, 1,
                  fError,
                  fErrorFile,
                  fErrorLine,
                  fErrorPos );

    ErrMessage( Format(CCompileError, [fError, fErrorFile,
     fErrorLine , FindLine( fErrorFile, fErrorLine )]) );

    if not ( csDesigning in ComponentState ) then
      ShowError;
  end;
end;

procedure TVirtComp.ShowError;
var
  Form : TForm;
  Memo : TMemo;
  SkipChars, I, J : Integer;
begin
  Form := TForm.Create( nil );
  Form.Left  := 50;
  Form.Top   := 200;
  Form.Width := 580;
  Memo := TMemo.Create( Form );
  Form.InsertControl( Memo );
  Form.Caption := Format(CFormCaption, [fError, fErrorFile]);

  if not ( csDesigning in ComponentState ) then
    Form.Caption := Form.Caption + CReadOnly;

  Memo.Align := alClient;

  if Assigned( Font ) then
    Memo.Font.Assign( Font )
  else
  begin
    Memo.Font.Name := 'Courier new';
    Memo.Font.Size := 12;
  end;

  try
    Memo.Lines.LoadFromFile( fErrorFile );

    Form.Show;

    I := fErrorLine;
    Skipchars := 0;
    for J :=1 to I - 1 do
      Skipchars := Skipchars + Length(Memo.Lines[J-1]) + 2;

    Memo.SetFocus;
    Memo.SelStart  := SkipChars + fErrorPos - 1;
    Memo.SelLength := 0;

    Form.Hide;

    if not ( csDesigning in ComponentState ) then
      Memo.ReadOnly := true;

    Form.ShowModal;

    if Memo.Modified then
    begin
      if MessageBox( GetActiveWindow(),
          PChar(Format(CSaveChanges, [fErrorFile])),
         PChar(CConfirm), MB_ICONQUESTION or MB_YESNO ) = IDYES then
        Memo.Lines.SaveToFile( fErrorFile );
    end;
  finally
    Form.Free;
  end;
end;

procedure TVirtComp.SetGlobals;
begin
  VirtSetRunComp( fH );
  RunVirtComp := Self;
end;

function TVirtComp.LoadScript : boolean;
begin
  if ScriptExists then
  begin
    Result := true;
    SetGlobals;
    if fState <> stLOAD then
    begin
      if UpperCase(ExtractFileExt( ScriptFileName )) = '.W' then
        Result := Parse
      else
        Result := LoadBinProg;
    end;
  end
  else
    Result := false;
end;

procedure TVirtComp.ProcessRunError;
begin
  if Assigned( fOnRunError ) then
     fOnRunError( Self )
  else
  begin
    VirtGetError( fH, 2,
                  fError,
                  fErrorFile,
                  fErrorLine,
                  fErrorPos );
     ShowError;
  end;
end;

function TVirtComp.Run : boolean;
begin
  Result := false;
  if LoadScript then
  begin
    Result := VirtRun( fH );
    if not Result then
      ProcessRunError;
  end;
end;

function TVirtComp.GetIDOfFunction( const FuncName : String ) : Integer;
var
  I : Integer;
begin
  I := fAssoc.IndexOf( FuncName );
  if I = -1 then
  begin
    Result := VirtGetIDOfFunction( FuncName );
    fAssoc.AddObject( FuncName, TObject( result ) );
  end
  else
    Result := Integer( fAssoc.Objects[I] );
end;

function TVirtComp.ExecFunction( const FuncName : String;
    const Params : array of Variant; var Res : Variant ) : boolean;
var
  I : Integer;
begin
  Result := false;
  if LoadScript then
  begin
    I := GetIDOfFunction( FuncName );
    if I > 0 then
      Result := VirtExecFunction( I, Params, Res )
    else
    begin
      ErrMessage( Format(CFunctionnotFound, [FuncName]) );
      Exit;
    end;
    if not Result then
      ProcessRunError;
  end;
end;

procedure TVirtComp.AppFunction( const FuncName : String;
                                 ParCount : Integer;
                                 Proc : TExternalFunc );
begin
  VirtAppFunction( fH, FuncName, ParCount, Proc );
end;

procedure TVirtComp.AppSystemFunction( const FuncName : String;
                                       ParCount : Integer;
                                       Proc : TExternalFunc );
begin
  VirtAppSystemFunction( fH, FuncName, ParCount, Proc );
end;

function  TVirtComp.GetIDOfVariable( const VarName : String ) : Integer;
var
  I : Integer;
begin
  I := fAssoc.IndexOf( VarName );
  if I = -1 then
  begin
    result := VirtGetIDOfVariable( VarName );
    fAssoc.AddObject( VarName, TObject( result ) );
  end
  else
    result := Integer( fAssoc.Objects[I] );
end;

function  TVirtComp.GetVar( const VarName : String; var Res : Variant ) : boolean;
var
  I : Integer;
begin
  I := GetIDOfVariable( VarName );
  if I > 0 then
  begin
    Res := VirtGetVar( I );
    result := true;
  end
  else
    result := false;
end;

function  TVirtComp.GetVarAsString( const VarName : String;
                                    var Res : ShortString ) : boolean;
var
  I : Integer;
begin
  I := GetIDOfVariable( VarName );
  if I > 0 then
  begin
    Res := VirtGetVarAsString( I );
    result := true;
  end
  else
    result := false;
end;

function TVirtComp.PutVarByID( ID : Integer; const Val : Variant ) : boolean;
begin
  VirtPutVar( ID, Val );
  result := true;
end;

function  TVirtComp.PutVar( const VarName : String; const Val : Variant ) : boolean;
var
  I : Integer;
begin
  I := GetIDOfVariable( VarName );
  if I > 0 then
  begin
    VirtPutVar( I, Val );
    result := true;
  end
  else
    result := false;
end;

procedure TVirtComp.SetVar( const VarName : String; const Val : Variant );
begin
  VirtSetVar( fH, VarName, Val );
end;

procedure  TVirtComp.GetLastError( var ErrorMessage, ErrorFile : ShortString;
                                   var ErrorLine, ErrorPos : Integer );
begin
  ErrorMessage := fError;
  ErrorFile    := fErrorFile;
  ErrorLine    := fErrorLine;
  ErrorPos     := fErrorPos;
end;

function TVirtComp.GetDllPath : String;
begin
  result := VirtGetDllPath;
end;

procedure TFileNameEdit.Edit;
begin
  with TOpenDialog.Create( nil ) do
    try
      Filter := 'VIRT files (*.w;*.wb)|*.W;*.WB' +
                '|VIRT source files (*.w)|*.W'   +
                '|VIRT binary files (*.wb)|*.WB';

      if Execute then
        SetStrValue( ExtractFileName( FileName ) );
    finally
      Free;
    end;
end;

function TFileNameEdit.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

function TFileNameEdit.GetValue : string;
begin
  result := GetStrValue;
end;

procedure TFileNameEdit.SetValue(const val : string);
begin
  SetStrValue( val );
end;

procedure TMemoEdit.Edit;
var
  V : TVirtComp;
  Form : TForm;
  Memo : TMemo;
  Panel: TPanel;
  C : TCheckBox;
  S : String;
  Successful : boolean;
begin
  V := GetComponent( 0 ) as TVirtComp;
  S := V.ScriptFileName;

  if UpperCase(ExtractFileExt( S )) = '.WB' then
  begin
    ErrMessage( CCannotView + S );
    Exit;
  end;

  Form := TForm.Create( nil );
  Form.Left  := 100;
  Form.Top   := 200;
  Form.Width := 500;

  Panel := TPanel.Create( Form );
  Form.InsertControl( Panel );
  Panel.Caption := '';
  Panel.Height := 20;
  Panel.BevelOuter := bvNone;
  Panel.Align := alBottom;

  C := TCheckBox.Create( Panel );
  C.Top  := 3;
  C.Left := 3;
  C.Width:= 130;
  C.Caption := 'Syntax checking';
  Panel.InsertControl( C );

  Memo := TMemo.Create( Form );
  Form.InsertControl( Memo );
  Form.Caption := 'Script editor';
  Memo.Align := alClient;

  if Assigned( V.Font ) then
    Memo.Font.Assign( V.Font )
  else
  begin
    Memo.Font.Name := 'Courier new';
    Memo.Font.Size := 12;
  end;

  try
    if FileExists( S ) then
      Memo.Lines.LoadFromFile( S );

    Form.ShowModal;

    if not FileExists( S ) then
      S := 'noname.w';

    S := ExtractFileName( S );

    V.fFileName := S;

    if Memo.Modified and not C.Checked then
    begin
      if UpperCase( S ) = 'NONAME.W' then
      begin
        V.fFileName := SaveNonameMemo( Memo );
      end
      else
      if MessageDlg(Format(CSaveChanges, [S]), mtConfirmation,
        [mbYes, mbNo], 0)=mrYes then
        Memo.Lines.SaveToFile( S );
    end;

    if C.Checked then
    begin
      if UpperCase( S ) = 'NONAME.W' then
        V.fFileName := SaveNonameMemo( Memo )
      else
        Memo.Lines.SaveToFile( S );

      V.OpenVirt;

      V.fModeParse := 1; // syntax checking only;

      Successful := V.Parse;
      V.CloseVirt;

      if not Successful then
        V.ShowError;

    end;
  finally
    Form.Free;
  end;
end;

function TMemoEdit.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

function TMemoEdit.GetValue : string;
begin
  result := {GetStrValue;} '(TStrings)';
end;

procedure TMemoEdit.SetValue(const val : string);
begin
  SetStrValue( val );
end;

{******************************************************************}

var
  ConsoleMemo : TMemo;
  ConsoleForm : TForm;

procedure _writeln; far;
var
  Q : String;
begin
  if FindWindow( 'TForm', 'VIRT console' ) = 0 then
  begin
    ConsoleForm := TForm.Create( RunVirtComp.Owner );
    ConsoleForm.Top  := 250;
    ConsoleForm.Left := 50;
    ConsoleForm.Width := 500;

    ConsoleMemo := TMemo.Create( ConsoleForm );
    ConsoleForm.InsertControl( ConsoleMemo );
    ConsoleForm.Caption := 'VIRT console';
//    ConsoleForm.FormStyle := fsStayOnTop;
    ConsoleMemo.Align := alClient;
    ConsoleMemo.Font.Name := 'Courier new';
    ConsoleMemo.Font.Size := 12;
    ConsoleMemo.Font.Color  := clWhite;
    ConsoleMemo.Brush.Color := clBlack;
    ConsoleMemo.ReadOnly := true;
    ConsoleForm.Show;
  end;

  Q := VirtPopVarAsString;
  ConsoleMemo.Lines.Add( Q );
  VirtPushVar( 0 );

//  if IsIconic( ConsoleForm.Handle ) then
//     OpenIcon( ConsoleForm.Handle );

end;

procedure TVirtCompEditor.Edit;
begin
  MessageDlg( 'VIRT language component, ver. 1.1' + #13#10 +
              'Copyrignt (c) 1999 by A.Baranovsky',
              mtInformation, [mbOK], 0 );
end;

procedure Register;
begin
  RegisterComponents( 'Samples', [TVirtComp]);
  RegisterPropertyEditor( TypeInfo( TFileName ), TVirtComp, '', TFileNameEdit);
  RegisterPropertyEditor( TypeInfo( TStrings ),  TVirtComp, '', TMemoEdit);
  RegisterComponentEditor( TVirtComp, TVirtCompEditor );
end;

initialization

  LibraryHandle := 0;
  VirtCount     := 0;

  UserLibraryCount := 0;
  UserFunctionCount := 0;

finalization

  if VirtCount > 0 then
  begin
    if LibraryHandle > 0 then
    begin
      VirtDone( VirtCount );
      FreeLibrary( LibraryHandle );
    end;

    while UserLibraryCount > 0 do
    begin
      if UserLibraryHandlers[ UserLibraryCount ] > 0 then
        FreeLibrary( UserLibraryHandlers[ UserLibraryCount ] );
      Dec( UserLibraryCount );
    end;
  end;

end.
